import os
from logging import Logger
from pathlib import Path
from typing import Tuple

from fs.base import FS
from fs.errors import DirectoryExpected, FileExpected
from fs.glob import imatch, match
from fs.memoryfs import MemoryFS
from fs.osfs import OSFS
from fs.path import join

from autoslug.utils.logging import log_access_denied


def _os_rename(fs: FS, old: str, new: str) -> bool:
    try:
        os.rename(src=fs.getospath(old), dst=fs.getospath(new))
    except PermissionError:
        return False
    return True


def _fs_rename(fs: FS, old: str, new: str) -> bool:
    try:
        if fs.isfile(old):
            fs.move(src_path=old, dst_path=new)
        else:
            fs.movedir(src_path=old, dst_path=new, create=True)
    except (FileExpected, DirectoryExpected):
        return False
    return True


def rename(fs: FS, old: str, new: str) -> bool:
    try:
        if fs.getmeta()["supports_rename"]:
            return _os_rename(fs=fs, old=old, new=new)
    except KeyError:
        pass
    return _fs_rename(fs=fs, old=old, new=new)


def match_globs(fs: FS, path: str, globs: Tuple[str]) -> bool:
    if fs.getmeta()["case_insensitive"]:
        return any([imatch(glob, path) for glob in globs])
    return any([match(glob, path) for glob in globs])


def _copy_structure(
    src_fs: FS, dst_fs: FS, src_path: str, dst_path: str, logger: Logger
) -> bool:
    ok = True
    if src_fs.isdir(src_path):
        dst_fs.makedirs(dst_path, recreate=True)
        try:
            for subpath in src_fs.scandir(src_path):
                ok = (
                    _copy_structure(
                        src_fs=src_fs,
                        dst_fs=dst_fs,
                        src_path=join(src_path, subpath.name),
                        dst_path=join(dst_path, subpath.name),
                        logger=logger,
                    )
                    and ok
                )
        except DirectoryExpected:
            log_access_denied(path=src_path, logger=logger)
            return False
    elif src_fs.isfile(src_path):
        dst_fs.create(dst_path)
    return ok


def get_filesystem(
    path: Path, ignore_root: bool, dry_run: bool, logger: Logger
) -> Tuple[FS, str, bool, bool]:
    ok = True
    if path.resolve().as_posix() == Path(os.getcwd()).resolve().as_posix():
        ignore_root = True
        root = path.as_posix()
        start = "/"
    else:
        root = path.parent.as_posix()
        start = path.name
    if dry_run:
        fs = MemoryFS()
        with OSFS(root) as osfs:
            ok = _copy_structure(
                src_fs=osfs, dst_fs=fs, src_path=start, dst_path=start, logger=logger
            )
    else:
        fs = OSFS(root)
    return fs, start, ignore_root, ok



import re
from logging import Logger
from typing import Dict, Optional, Set, Tuple

from fs.base import FS
from fs.errors import DirectoryExpected
from fs.path import basename, dirname, join, splitext
from inflection import dasherize, parameterize, underscore
from slugify import SLUG_OK, slugify

from autoslug.utils.filesystem import match_globs, rename
from autoslug.utils.logging import log_access_denied, log_ignored


def _handle_affixes(
    stem: str, prefixes: Set[str], suffixes: Set[str]
) -> Tuple[str, str, str]:
    prefix_pattern = "|".join(re.escape(prefix) + "+" for prefix in prefixes)
    suffix_pattern = "|".join(re.escape(suffix) + "+" for suffix in suffixes)
    pattern = f"^({prefix_pattern})?(.+?)({suffix_pattern})?$"
    match = re.match(pattern, stem)
    if match:
        prefix = match.group(1) or ""
        stem = match.group(2) or ""
        suffix = match.group(3) or ""
    else:
        prefix = ""
        suffix = ""
    return prefix, stem, suffix


def _shorten_stem(stem: str, max_length: Optional[int], sep: str) -> str:
    if len(stem) <= max_length:
        return stem
    parts = stem.split(sep)
    new_stem = parts.pop(0)
    for part in parts:
        if len(new_stem) + len(sep) + len(part) > max_length:
            break
        new_stem += sep + part
    return new_stem


def _extract_leading_digits(stem: str, sep: str, n: Optional[int]) -> Tuple[str, str]:
    if n is not None:
        parts = stem.split(sep)
        try:
            if parts[0].isdigit() and parts[1].isalpha():
                number = str(min(int(parts[0]), 10**n - 1)).zfill(n)
                return number, sep.join(parts[1:])
        except IndexError:
            pass
    return "", stem


def _process_stem(
    stem: str,
    dash: bool,
    prefixes: Set[str],
    suffixes: Set[str],
    max_length: Optional[int],
    n_digits: Optional[int],
) -> str:
    prefix, stem, suffix = _handle_affixes(
        stem=stem, prefixes=prefixes, suffixes=suffixes
    )
    stem = parameterize(
        slugify(
            s=underscore(stem),
            ok=(SLUG_OK + "."),
            only_ascii=True,
        )
    )
    stem = dasherize(stem) if dash else underscore(stem)
    sep = "-" if dash else "_"
    stem = re.sub(f"{sep}+", sep, stem).strip(sep)
    digits, stem = _extract_leading_digits(stem=stem, sep=sep, n=n_digits)
    if max_length is not None:
        if prefix is not None:
            max_length -= len(prefix)
            if len(digits) > 0:
                max_length -= len(digits) + len(sep)
        stem = _shorten_stem(stem=stem, max_length=max_length, sep=sep)
    return prefix + (digits + sep if len(digits) > 0 else "") + stem + suffix


def _process_ext(ext: str, mappings: Dict[str, str]) -> str:
    try:
        return f".{mappings[ext.strip('.')]}"
    except KeyError:
        return ext


def _check_conflict(fs: FS, path: str, new_path: str) -> bool:
    try:
        if fs.getmeta()["case_insensitive"]:
            if path.lower() == new_path.lower():
                return False
    except KeyError:
        pass
    return fs.exists(new_path)


def _process_change(
    fs: FS,
    path: str,
    new_path: str,
    logger: Logger,
    warn_limit: Optional[int],
    error_limit: Optional[int],
) -> bool:
    change = path != new_path
    new_path_len = len(new_path)
    if change:
        if _check_conflict(fs=fs, path=path, new_path=new_path):
            logger.error(f"conflict preventing renaming: {path} -> {new_path}")
        else:
            if rename(fs=fs, old=path, new=new_path):
                logger.info(f"renamed: {path} -> {new_path}")
            else:
                log_access_denied(path=path, logger=logger)
                return False
    else:
        logger.debug(f"unchanged: {path}")
    if error_limit is not None:
        if new_path_len > error_limit:
            logger.error(f"error (path exceeds {error_limit} characters): {new_path}")
        return False
    if warn_limit is not None:
        logger.warning(f"warning (path exceeds {warn_limit} characters): {new_path}")
    return not change


def _process_file(
    fs: FS,
    path: str,
    ok_exts: Set[str],
    no_dash_exts: Set[str],
    ext_map: Dict[str, str],
    prefixes: Set[str],
    suffixes: Set[str],
    logger: Logger,
    warn_limit: Optional[int],
    error_limit: Optional[int],
    max_length: Optional[int],
    n_digits: Optional[int],
) -> bool:
    suffix = splitext(path)[1]
    if suffix in ok_exts:
        stem = splitext(basename(path))[0]
    else:
        stem = basename(path)
        suffix = ""
    dash = suffix not in no_dash_exts
    new_path = join(
        dirname(path),
        _process_stem(
            stem=stem,
            dash=dash,
            prefixes=prefixes,
            suffixes=suffixes,
            max_length=max_length,
            n_digits=n_digits,
        )
        + _process_ext(ext=suffix, mappings=ext_map),
    )
    return _process_change(
        fs=fs,
        path=path,
        new_path=new_path,
        logger=logger,
        warn_limit=warn_limit,
        error_limit=error_limit,
    )


def _process_dir(
    fs: FS,
    path: str,
    ignore_globs: Set[str],
    ok_exts: Set[str],
    no_dash_exts: Set[str],
    ext_map: Dict[str, str],
    prefixes: Set[str],
    suffixes: Set[str],
    ignore_root: bool,
    no_recurse: bool,
    logger: Logger,
    warn_limit: Optional[int],
    error_limit: Optional[int],
    max_length: Optional[int],
    n_digits: Optional[int],
) -> bool:
    ok = True
    if not ignore_root:
        new_path = join(
            dirname(path),
            _process_stem(
                stem=basename(path),
                dash=True,
                prefixes=prefixes,
                suffixes=suffixes,
                max_length=max_length,
                n_digits=n_digits,
            ),
        )
        ok = (
            _process_change(
                fs=fs,
                path=path,
                new_path=new_path,
                logger=logger,
                warn_limit=warn_limit,
                error_limit=error_limit,
            )
            and ok
        )
        path = new_path
    if not no_recurse:
        try:
            for subpath in fs.scandir(path):
                ok = (
                    process_path(
                        fs=fs,
                        path=join(path, subpath.name),
                        ignore_globs=ignore_globs,
                        ok_exts=ok_exts,
                        no_dash_exts=no_dash_exts,
                        ext_map=ext_map,
                        prefixes=prefixes,
                        suffixes=suffixes,
                        ignore_root=False,
                        no_recurse=False,
                        logger=logger,
                        warn_limit=warn_limit,
                        error_limit=error_limit,
                        max_length=max_length,
                        n_digits=n_digits,
                    )
                    and ok
                )
        except DirectoryExpected:
            log_access_denied(path=path, logger=logger)
            return False
    else:
        log_ignored(path=path, logger=logger)
    return ok


def process_path(
    fs: FS,
    path: str,
    ignore_globs: Set[str],
    ok_exts: Set[str],
    no_dash_exts: Set[str],
    ext_map: Dict[str, str],
    prefixes: Set[str],
    suffixes: Set[str],
    ignore_root: bool,
    no_recurse: bool,
    logger: Logger,
    warn_limit: Optional[int],
    error_limit: Optional[int],
    max_length: Optional[int],
    n_digits: Optional[int],
) -> bool:
    if match_globs(fs=fs, path=path, globs=ignore_globs):
        log_ignored(path=path, logger=logger)
        return True
    elif fs.isdir(path):
        return _process_dir(
            fs=fs,
            path=path,
            ignore_globs=ignore_globs,
            ok_exts=ok_exts,
            no_dash_exts=no_dash_exts,
            ext_map=ext_map,
            prefixes=prefixes,
            suffixes=suffixes,
            ignore_root=ignore_root,
            no_recurse=no_recurse,
            logger=logger,
            warn_limit=warn_limit,
            error_limit=error_limit,
            max_length=max_length,
            n_digits=n_digits,
        )
    elif fs.isfile(path):
        return _process_file(
            fs=fs,
            path=path,
            ok_exts=ok_exts,
            no_dash_exts=no_dash_exts,
            ext_map=ext_map,
            prefixes=prefixes,
            suffixes=suffixes,
            logger=logger,
            warn_limit=warn_limit,
            error_limit=error_limit,
            max_length=max_length,
            n_digits=n_digits,
        )
    else:
        logger.debug(f"skipped (not a file or directory): {path}")
        return True



from mimetypes import types_map
from pathlib import Path
from typing import Set

from autoslug.__version__ import __version__
from autoslug.config.defaults import (
    IGNORE_GLOBS,
    NO_DASH_EXTS,
    OK_EXTS,
    PREFIXES,
    SUFFIXES,
)


def add_mime_types(exts: Set[str]) -> Set[str]:
    return exts.union(set(types_map.keys()))


POSITIONAL = [
    {
        "path": {
            "help": "path to the file or directory to process",
            "metavar": "<path>",
            "type": str,
            "postprocess": [lambda x: Path(x).resolve()],
        }
    }
]

OPTIONAL = {
    "dry_run": {
        "action": "store_true",
        "help": "do not actually rename files or directories",
        "shorthands": ["d", "n"],
    },
    "error_limit": {
        "default": None,
        "help": (
            "set exit level to failure if any paths exceed this character limit "
            "(will still attempt to process all paths)"
        ),
        "metavar": "<n>",
        "type": int,
    },
    "force": {
        "action": "store_true",
        "help": "disable protections and force processing",
        "shorthands": ["f"],
    },
    "ignore_globs": {
        "action": "extend",
        "default": list(IGNORE_GLOBS),
        "help": "glob patterns to ignore",
        "metavar": "<glob>",
        "nargs": "*",
        "postprocess": [set],
        "type": str,
    },
    "ignore_root": {
        "action": "store_true",
        "help": (
            "only process children of the specified path "
            "(implied when running in current directory)"
        ),
    },
    "log_file": {
        "default": None,
        "help": (
            "log output to specified file "
            "(all messages logged to file regardless of log level)"
        ),
        "metavar": "<path>",
        "type": str,
    },
    "log_level": {
        "choices": ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        "default": "INFO",
        "help": (
            "set the console logging level to one of "
            "DEBUG, INFO (default), WARNING, ERROR, or CRITICAL "
            "(overrides --quiet and --verbose)"
        ),
        "metavar": "<level>",
        "type": str,
    },
    "max_length": {
        "default": None,
        "help": (
            "attempt to shorten file and directory names "
            "to not exceed this number of characters (excluding extension)"
        ),
        "metavar": "<n>",
        "type": int,
    },
    "no_dash_exts": {
        "action": "extend",
        "default": list(NO_DASH_EXTS),
        "help": (
            "file extensions (without periods) "
            "where underscores should be used instead of dashes"
        ),
        "metavar": "<ext>",
        "nargs": "*",
        "postprocess": [set, lambda x: {f".{ext}" for ext in x}],
        "type": str,
    },
    "no_recurse": {
        "action": "store_true",
        "help": "do not recurse into subdirectories (only process specified path)",
    },
    "num_digits": {
        "default": None,
        "help": (
            "attempt to pad or round any existing numerical prefixes "
            "to consist of this many digits"
        ),
        "metavar": "<n>",
        "type": int,
    },
    "ok_exts": {
        "action": "extend",
        "default": list(OK_EXTS),
        "help": "file extensions (without periods) to recognize",
        "help_suffix": (
            "and common MIME types "
            "(all other extensions are treated as part of the filename)"
        ),
        "metavar": "<ext>",
        "nargs": "*",
        "postprocess": [
            set,
            lambda x: {f".{ext}" for ext in x},
            lambda x: add_mime_types(x),
        ],
        "type": str,
    },
    "prefixes": {
        "action": "extend",
        "default": list(PREFIXES),
        "help": "file or directory name prefixes to leave unchanged",
        "metavar": "<prefix>",
        "nargs": "*",
        "postprocess": [set],
        "type": str,
    },
    "quiet": {
        "action": "store_true",
        "help": (
            "suppress all output except errors "
            "(equivalent to setting --log-level=ERROR)"
        ),
        "shorthands": ["q"],
    },
    "suffixes": {
        "action": "extend",
        "default": list(SUFFIXES),
        "help": "file or directory name suffices (before extension) to leave unchanged",
        "metavar": "<suffix>",
        "nargs": "*",
        "postprocess": [set],
        "type": str,
    },
    "verbose": {
        "action": "store_true",
        "help": (
            "report skipped and ignored paths in addition to renamed ones"
            "(equivalent to setting --log-level=DEBUG, overrides --quiet)"
        ),
        "shorthands": ["v"],
    },
    "version": {
        "action": "version",
        "help": "display version information and exit",
        "version": f"%(prog)s {__version__}",
    },
    "warn_limit": {
        "default": None,
        "help": (
            "output warning if path exceeds this character limit "
            "(does not affect exit level)"
        ),
        "metavar": "<n>",
        "type": int,
    },
}



from logging import (
    DEBUG,
    INFO,
    FileHandler,
    Formatter,
    Logger,
    StreamHandler,
    getLogger,
)

from autoslug.config.defaults import (
    LOG_CONSOLE_FORMAT,
    LOG_DATE_FORMAT,
    LOG_FILE_FORMAT,
)


def get_logger(
    name: str = "autoslug",
    level: int = DEBUG,
    console_level: int = INFO,
    file_level=DEBUG,
    log_file: str = None,
) -> Logger:
    logger = getLogger(name)
    logger.setLevel(level)
    console_handler = StreamHandler()
    console_handler.setLevel(console_level)
    console_handler.setFormatter(Formatter(LOG_CONSOLE_FORMAT, datefmt=LOG_DATE_FORMAT))
    logger.addHandler(console_handler)
    if log_file:
        file_handler = FileHandler(log_file)
        file_handler.setLevel(file_level)
        file_handler.setFormatter(Formatter(LOG_FILE_FORMAT, datefmt=LOG_DATE_FORMAT))
        logger.addHandler(file_handler)
    return logger


def log_access_denied(path: str, logger: Logger) -> None:
    logger.error(f"access denied: {path}")
    return None


def log_ignored(path: str, logger: Logger) -> None:
    logger.debug(f"ignored: {path}")
    return None


def log_ignored(path: str, logger: Logger) -> None:
    logger.debug(f"ignored: {path}")
    return None



from argparse import ArgumentParser
from typing import Any, Dict, List, Optional


def _extend_help_text(
    message: str, defaults: List[str], suffix: Optional[str] = None
) -> str:
    if not defaults:
        return message
    defaults = sorted(defaults)
    text = message + " in addition to "
    if len(defaults) >= 2 and suffix is not None:
        text += '"' + '", "'.join(defaults) + '", and ' + suffix
    elif len(defaults) > 2:
        text += '"' + '", "'.join(defaults[:-1]) + '", and "' + defaults[-1] + '"'
    elif len(defaults) == 2:
        text += f'"{defaults[0]}" and "{defaults[1]}"'
    elif suffix is not None:
        text += '"' + defaults[0] + '" and ' + suffix
    else:
        text += '"' + defaults[0] + '"'
    return text


def _process_positional(
    parser: ArgumentParser, args: List[Dict[str, Dict[str, Any]]]
) -> None:
    for arg in args:
        name = list(arg.keys())[0]
        params = list(arg.values())[0]
        params = params.copy()
        params.pop("postprocess", None)
        parser.add_argument(name.replace("_", "-"), **params)
        return None


def _get_order(args: Dict[str, Dict[str, Any]]) -> List[str]:
    with_sh = []
    min_sh = []
    without_sh = []
    for name, params in args.items():
        shorthands = params.get("shorthands", [])
        if shorthands:
            with_sh.append(name)
            min_sh.append(min(shorthands))
        else:
            without_sh.append(name)
    return [name for _, name in sorted(zip(min_sh, with_sh))] + sorted(without_sh)


def _process_optional(parser: ArgumentParser, args: Dict[str, Dict[str, Any]]) -> None:
    for name in _get_order(args=args):
        params = args[name].copy()
        shorthands = [f"-{shorthand}" for shorthand in params.pop("shorthands", [])]
        names_flags = sorted(shorthands) + [f"--{name.replace('_', '-')}"]
        try:
            if params["action"] == "extend":
                params["help"] = _extend_help_text(
                    message=params["help"],
                    defaults=params["default"],
                    suffix=params.pop("help_suffix", None),
                )
        except KeyError:
            pass
        params.pop("postprocess", None)
        parser.add_argument(*names_flags, **params)
    return None


def _postprocess(
    args: Dict[str, Any], params: Dict[str, Dict[str, Any]]
) -> Dict[str, Any]:
    for name, value in args.items():
        for func in params[name].get("postprocess", []):
            value = func(value)
        args[name] = value
    return args


def parse_arguments(
    description: str,
    positional: List[Dict[str, Dict[str, Any]]],
    optional: Dict[str, Dict[str, Any]],
) -> Dict[str, Any]:
    parser = ArgumentParser(description=description)
    _process_positional(parser=parser, args=positional)
    _process_optional(parser=parser, args=optional)
    return _postprocess(
        args=vars(parser.parse_args()),
        params=(
            {list(arg.keys())[0]: list(arg.values())[0] for arg in positional}
            | optional
        ),
    )



from typing import Dict, Final, Set

EXT_MAP: Final[Dict[str, str]] = {
    "yml": "yaml",
}
IGNORE_GLOBS: Final[Set[str]] = {
    "**/__pycache__",
    "**/.DS_Store",
    "**/.git",
    "**/LICENSE*",
    "**/README*",
}
NO_DASH_EXTS: Final[Set[str]] = {
    "py",
}
OK_EXTS: Final[Set[str]] = {
    "cmd",
    "ipynb",
    "md",
    "ps1",
    "R",
    "Rmd",
    "rst",
    "yaml",
    "yml",
}
PREFIXES: Final[Set[str]] = {
    "_",
    ".",
}
SUFFIXES: Final[Set[str]] = {
    "_",
}

DESCRIPTION: Final[str] = (
    "automatically rename files and directories to be URL-friendly"
)

LOG_CONSOLE_FORMAT: Final[str] = "[%(levelname)s] %(message)s"
LOG_DATE_FORMAT: Final[str] = "%Y-%m-%d %H:%M:%S"
LOG_FILE_FORMAT: Final[str] = "%(asctime)s [%(levelname)s] %(message)s"



import logging
from pathlib import Path
from subprocess import DEVNULL, CalledProcessError, run
from typing import Optional, Tuple

from autoslug.autoslug import process_path
from autoslug.config.arguments import OPTIONAL, POSITIONAL
from autoslug.config.defaults import DESCRIPTION, EXT_MAP
from autoslug.utils.filesystem import get_filesystem
from autoslug.utils.logging import get_logger
from autoslug.utils.parser import parse_arguments


def get_log_level(level: str, quiet: bool, verbose: bool) -> int:
    if level != "INFO":
        return getattr(logging, level)
    elif verbose:
        return logging.DEBUG
    elif quiet:
        return logging.ERROR
    else:
        return getattr(logging, "INFO")


def is_git_repository(path: Path) -> Tuple[bool, Optional[bool]]:
    try:
        run(
            ["git", "-C", path.as_posix(), "rev-parse"],
            check=True,
            stdout=DEVNULL,
            stderr=DEVNULL,
        )
        return True, True
    except CalledProcessError:
        return True, False
    except FileNotFoundError:
        return False, None


def check_git_repository(path: Path, force: bool, logger: logging.Logger) -> None:
    test_ok, is_git = is_git_repository(path=path)
    if not test_ok:
        msg = "unable to determine whether path is within git repository"
    elif not is_git:
        msg = "specified path is not within a git repository"
    if (not test_ok or not is_git) and not force:
        logger.critical(f"{msg}: {path.as_posix()}")
        logger.warning("actions might be destructive and irreversible")
        logger.info("run again with --force to override and process anyway")
        exit(1)
    return None


def assert_path(path: Path, logger: logging.Logger) -> None:
    if not path.exists():
        logger.critical(f"specified path does not exist: {path.as_posix()}")
        exit(1)
    return None


def perform_checks(path: Path, force: bool, logger: logging.Logger) -> None:
    assert_path(path=path, logger=logger)
    check_git_repository(path=path, force=force, logger=logger)
    return None


def main() -> None:

    args = parse_arguments(
        description=DESCRIPTION, positional=POSITIONAL, optional=OPTIONAL
    )

    logger = get_logger(
        console_level=get_log_level(
            level=args["log_level"], quiet=args["quiet"], verbose=args["verbose"]
        ),
        log_file=args["log_file"],
    )

    perform_checks(path=args["path"], force=args["force"], logger=logger)

    if args["force"]:
        logger.warning("disabling protections and forcing processing")
    if args["dry_run"]:
        logger.info("beginning dry run (no changes will be made)")

    fs, start, ignore_root, ok = get_filesystem(
        path=args["path"],
        ignore_root=args["ignore_root"],
        dry_run=args["dry_run"],
        logger=logger,
    )

    ok = (
        process_path(
            error_limit=args["error_limit"],
            ext_map=EXT_MAP,
            fs=fs,
            ignore_globs=args["ignore_globs"],
            ignore_root=ignore_root,
            logger=logger,
            max_length=args["max_length"],
            n_digits=args["num_digits"],
            no_dash_exts=args["no_dash_exts"],
            no_recurse=args["no_recurse"],
            ok_exts=args["ok_exts"],
            path=start,
            prefixes=args["prefixes"],
            suffixes=args["suffixes"],
            warn_limit=args["warn_limit"],
        )
        and ok
    )

    fs.close()

    if not ok:
        exit(1)


if __name__ == "__main__":
    main()
